////////////////////////////////////////////////////////////////////// 
// FastFit.cpp 
// (c) 2012-2013 Shuang Feng, Dajiang Liu, Goncalo Abecasis
// 
// This file is distributed as part of the RareMetalWorker source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile RareMetalWorker.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday November 28, 2012
// 

#include "FastFit.h"
#include "PreMeta.h"

String FastFit::traitName = "";
bool FastFit::useCovariates = false;
bool FastFit::inverseNormal = false;
bool FastFit::regressOutCov = false;
bool FastFit::unrelated = false;

void FastFit::OptimizeDelta(double tol,FastTransform & trans)
{
   double inc = 0.2;
   for(int i=0;i<100;i++)
   {
      //printf("Start optimization ...\n");
      a = exp(-10.0+i*inc); fa=-1.0*Evaluate(a,trans,false);
      c = exp(-10.0+(i+1)*inc); fc=-1.0*Evaluate(c,trans,false);
      b = (a+c)/2.0; fb=-1.0*Evaluate(b,trans,false);

      double ll = -Brent(tol,trans);
      if(i==0) 
      {
	 logLikelihood = ll;
	 deltaHat = min;
      }
      if(ll>logLikelihood) 
      {
	 logLikelihood = ll;
	 deltaHat = min;
      }
//printf("logLikelihood is:%g,deltaHat is:%g\n",logLikelihood,deltaHat);
   }
}

void FastFit::GetBeta(double delta,FastTransform & trans)
{
   Matrix A;
   Vector B;
   A.Dimension(trans.fixedEffects,trans.fixedEffects,0.0);
   B.Dimension(trans.fixedEffects);
   B.Zero();

   //Now fill in matrix A
   Matrix subUXXU;
   subUXXU.Dimension(trans.fixedEffects,trans.fixedEffects);
   subUXXU.Zero();
   //fill in the subUXXU matrix
   for(int i=0;i<trans.persons;i++)
   {
      for(int r=0;r<trans.fixedEffects;r++) 
      {
	 for(int c=0;c<trans.fixedEffects;c++)
	 {
	    if(r==c) {
	       subUXXU[r][c] = trans.UX[i][c]*trans.UX[i][c];
	       subUXXU[r][c]= subUXXU[r][c]/(trans.D[i]+delta);
	    } 
	    else 
	    {
	       subUXXU[r][c]= trans.UX[i][r]*trans.UX[i][c];
	       subUXXU[r][c]= subUXXU[r][c]/(trans.D[i]+delta);
	       subUXXU[c][r]=subUXXU[r][c];
	    }
	 }
      }
      A.Add(subUXXU);
   }

   SVD svd; 
   svd.InvertInPlace(A);

   //Now fill in vector B
   Vector tmpAddToA;
   for(int i=0;i<trans.persons;i++)
   {
      tmpAddToA.Copy(trans.UX[i]);
      double factor = trans.UY[i]/(trans.D[i]+delta);
      tmpAddToA.Multiply(factor);
      B.Add(tmpAddToA);
   }
   beta.Product(A,B);
}

double FastFit::GetSigma_g(double delta,Vector & beta,FastTransform & trans)
{
   double sigma =0.0;
   for(int i=0;i<trans.persons;i++)
   {
      double factor = trans.UY[i] - trans.UX[i].InnerProduct(beta);
      sigma += factor*factor/(trans.D[i]+delta);
   }
   sigma /=trans.persons;
   return sigma;
}

double FastFit::Evaluate(double delta,FastTransform & trans,bool full)
{
   double logLikelihood;
   //Fill up beta vector
   GetBeta(delta,trans);	
   double sigma = GetSigma_g(delta, beta,trans);
   //calculate the constant
   double constant =0.0;
   for(int i=0;i<trans.persons;i++)
      constant += log(trans.D[i] + delta);
   if(full)
      logLikelihood = -0.5*(trans.persons*log(2*M_PI)+constant + trans.persons + trans.persons*log(sigma));
   else logLikelihood = -0.5*(constant + trans.persons + trans.persons*log(sigma));
   return logLikelihood;
}

double FastFit::EvaluateMulti(FastTransform & trans,bool full)
{
   double logLikelihood;
   double constant =0.0;
   for(int i=0;i<trans.persons;i++)
      constant += log(trans.D[i]);

   Vector residuals,fixedEff,weightedRes;

   fixedEff.Product(trans.UX,betaHat);
   residuals.Copy(trans.UY);
   residuals.Subtract(fixedEff);
   weightedRes.Copy(residuals);

   for(int i=0;i<trans.persons;i++)
      weightedRes[i] /= trans.D[i];

   if(full)
      logLikelihood = -0.5*(trans.persons*log(2*M_PI)+ constant + weightedRes.InnerProduct(residuals));
   else
   {
      logLikelihood = -0.5*(constant + weightedRes.InnerProduct(residuals));
   } 
   return logLikelihood;
}

void FastFit::PrintPreface(FILE * output,Pedigree & ped,FastTransform & trans,String & traitName,FILE * log)
{
   fprintf(output,"##Version=%s\n",VERSION);
   fprintf(output,"##Samples=%d\n",trans.totalN);
   fprintf(output,"##AnalyzedSamples=%d\n",trans.persons);
   fprintf(output,"##Families=%d\n",trans.numFamily);
   fprintf(output,"##AnalyzedFamilies=%d\n",trans.families);
   fprintf(output,"##Founders=%d\n",trans.numFounder);
   fprintf(output,"##AnalyzedFounders=%d\n",trans.analyzedFounders);

   if(useCovariates || regressOutCov)
   {
      fprintf(output,"##Covariates=");
      fprintf(output,"%s",ped.covariateNames[0].c_str());
      for(int i=1;i<ped.covariateCount;i++)
	 fprintf(output,",%s",ped.covariateNames[i].c_str());
      fprintf(output,"\n");
      fprintf(output,"##CovariateSummaries\tmin\t25th\tmedian\t75th\tmax\tmean\tvariance\n");
      for(int i=0;i<ped.covariateCount;i++)
      {
	 Vector covariate;
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for(int j=0;j<trans.pPheno[f].Length();j++)
	    {
	       covariate.Push(ped[trans.pPheno[f][j]].covariates[i]);
	    }
	 }
	 covariate.Sort();
	 //covariate.RemoveDuplicates();
	 int first,second,middle;
	 first = (int) covariate.Length()/4;
	 second = first*3;
	 middle = (int) (first+second)/2;
	 fprintf(output,"##%s\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",ped.covariateNames[i].c_str(),covariate.Min(),covariate[first],covariate[middle],covariate[second],covariate.Max(),covariate.Average(),covariate.Var());
      }
   }
   fprintf(output,"##InverseNormal=%s\n",inverseNormal?"ON":"OFF");
   int traitNum = GetTraitID(ped,traitName.c_str(),log);
   fprintf(output,"##TraitSummaries\tmin\t25th\tmedian\t75th\tmax\tmean\tvariance\n");
   Vector y;
   for (int f = 0; f < ped.familyCount; f++) 
   {    
      for(int j=0;j<trans.pPheno[f].Length();j++)
      {
	 y.Push(ped[trans.pPheno[f][j]].traits[traitNum]);
      }         
   }
   y.Sort();
   int first,second,middle;
   first = (int) y.Length()/4;
   second = first*3;
   middle = (int) (first+second)/2;
   fprintf(output,"##%s\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",traitName.c_str(),y.Min(),y[first],y[middle],y[second],y.Max(),y.Average(),y.Var());
}

void FastFit::FitLinearModels(FILE * SCOREoutput,AutoFit & fitEngine, Pedigree & ped,double tol,FastTransform & trans,String & traitName,FILE * log)
{
   printf("Start fitting linear model for trait %s ...\n",traitName.c_str());
   fprintf(log,"Start fitting linear model for trait %s ...\n",traitName.c_str());
   FastFit::traitName = traitName;
   int traitNum = GetTraitID(ped,traitName.c_str(),log);
   trans.GetFixedEffectsCount(ped,useCovariates);

   trans.GetFixedEffectsCount(ped,useCovariates);
   beta.Dimension(trans.fixedEffects);
   betaHat.Dimension(trans.fixedEffects);

   trans.Prepare(ped,traitNum,useCovariates,log);

   PrintPreface(SCOREoutput,ped,trans,traitName,log);

   //if dealing with unrelated individuals
   if(unrelated)
   {
      //if --inverseNormal only
      if(inverseNormal && !regressOutCov)
      {
	 InverseNormalTransform(ped,traitNum);
	 Matrix inv;
	 Matrix transX;
	 transX.Transpose(trans.X);

	 Vector Y;
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for(int j=0;j<trans.pPheno[f].Length();j++)
	    {
	       Y.Push(ped[trans.pPheno[f][j]].traits[traitNum]);
	    }
	 }

	 Vector tmp;
	 tmp.Dimension(trans.fixedEffects);
	 for(int j=0;j<trans.fixedEffects;j++)
	    tmp[j] =Y.InnerProduct(transX[j]);

	 inv.Product(transX,trans.X);
	 SVD svd;
	 svd.InvertInPlace(inv);

	 for(int i=0;i<trans.fixedEffects;i++)
	    betaHat[i] = inv[i].InnerProduct(tmp);

	 residuals_unrelated.Dimension(trans.persons);
	 for(int i=0;i<trans.persons;i++)
	    residuals_unrelated[i] = Y[i] - betaHat.InnerProduct(trans.X[i]);

//         for(int i=0;i<trans.persons;i++)
//printf("%g\n",residuals_unrelated[i]);
	 sigma_g2Hat=0.0;
	 sigma2Hat = residuals_unrelated.SumSquares()/trans.persons;

	 Y.Sort();
	 int first,second,middle;
	 first = (int) Y.Length()/4;
	 second = first*3;
	 middle = (int) (first+second)/2;
	 fprintf(SCOREoutput,"##AnalyzedTrait\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",Y.Min(),Y[first],Y[middle],Y[second],Y.Max(),Y.Average(),Y.Var());
      }
      //if --inverseNormal --makeResiduals
      if(inverseNormal && regressOutCov)
      {
	 Matrix inv;
	 Matrix transX;
	 transX.Transpose(trans.X);

	 //Get beta_hat and residuals 
	 Vector tmp;
	 tmp.Dimension(trans.fixedEffects);
	 for(int j=0;j<trans.fixedEffects;j++)
	    tmp[j] =trans.Y.InnerProduct(transX[j]);

	 inv.Product(transX,trans.X);
	 SVD svd;
	 svd.InvertInPlace(inv);

	 for(int i=0;i<trans.fixedEffects;i++)
	    betaHat[i] = inv[i].InnerProduct(tmp);

	 residuals_unrelated.Dimension(trans.persons);
	 for(int i=0;i<trans.persons;i++)
	    residuals_unrelated[i] = trans.Y[i] - betaHat.InnerProduct(trans.X[i]);
	 //replace trait values into residuals
	 int index = 0;
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
	    {
	       if (ped[i].isPhenotyped(traitNum) && (!useCovariates || ped[i].isFullyControlled()))
	       {
		  if(PreMeta::genoFromVCF)
		  {
		     int p = trans.sampleVCFIDHash.Integer(ped[i].pid);
		     if(p!=-1)
		     {
			ped[i].traits[traitNum] = residuals_unrelated[index];
			index++;
		     }
		  }
		  if(PreMeta::genoFromPed)
		  {
		     int p = trans.samplePEDIDHash.Integer(ped[i].pid);
		     if(p!=-1)
		     {
			ped[i].traits[traitNum] = residuals_unrelated[index];
			index++;
		     }
		  }
	       }
	    }
	 }

	 InverseNormalTransform(ped,traitNum);

	 index=0;
	 Vector Y;
	 Y.Dimension(trans.persons);
	 for (int f = 0; f < ped.familyCount; f++) 
	 {
	    int ct = trans.pPheno[f].Length();
	    for(int i=0;i<ct;i++)
	    {
	       residuals_unrelated[index]=ped[trans.pPheno[f][i]].traits[traitNum];	 
	       Y[index] = residuals_unrelated[index];
	       index++;
	    }
	 }

	 Y.Sort();
	 int first,second,middle;
	 first = (int) Y.Length()/4;
	 second = first*3;
	 middle = (int) (first+second)/2;
	 fprintf(SCOREoutput,"##AnalyzedTrait\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",Y.Min(),Y[first],Y[middle],Y[second],Y.Max(),Y.Average(),Y.Var());

	 trans.traitMean = residuals_unrelated.Average();
	 trans.traitVar = residuals_unrelated.Var();
	 sigma_g2Hat=0.0;
	 sigma2Hat = residuals_unrelated.SumSquares()/trans.persons;
      }

      if(!inverseNormal)
      {
	 Matrix inv;
	 Matrix transX;
	 transX.Transpose(trans.X);

	 //Get beta_hat and residuals 
	 Vector tmp;
	 tmp.Dimension(trans.fixedEffects);
	 for(int j=0;j<trans.fixedEffects;j++)
	    tmp[j] =trans.Y.InnerProduct(transX[j]);

	 inv.Product(transX,trans.X);
	 SVD svd;
	 svd.InvertInPlace(inv);

	 for(int i=0;i<trans.fixedEffects;i++)
	    betaHat[i] = inv[i].InnerProduct(tmp);

	 residuals_unrelated.Dimension(trans.persons);
	 for(int i=0;i<trans.persons;i++)
	    residuals_unrelated[i] = trans.Y[i] - betaHat.InnerProduct(trans.X[i]);

	 Vector Y;
	 Y.Copy(trans.Y);
	 Y.Sort();
	 int first,second,middle;
	 first = (int) Y.Length()/4;
	 second = first*3;
	 middle = (int) (first+second)/2;
	 fprintf(SCOREoutput,"##AnalyzedTrait\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",Y.Min(),Y[first],Y[middle],Y[second],Y.Max(),Y.Average(),Y.Var());

	 trans.traitMean = residuals_unrelated.Average();
	 trans.traitVar = residuals_unrelated.Var();
	 sigma_g2Hat=0.0;
	 sigma2Hat = residuals_unrelated.SumSquares()/trans.persons;
      }
   }
   else
   {
      if(!FastTransform::empKin && !FastTransform::pedKin)
      {
	 FastTransform::pedKin  = true;
      }
      //ignore variance components and fit linear regression model
      if(regressOutCov)
      {
	 //fill in X and Y

	 Matrix inv;
	 Matrix transX;
	 transX.Transpose(trans.X);

	 //Get beta_hat and residuals 
	 Vector tmp;
	 tmp.Dimension(trans.fixedEffects);
	 for(int j=0;j<trans.fixedEffects;j++)
	    tmp[j] =trans.Y.InnerProduct(transX[j]);

	 inv.Product(transX,trans.X);
	 SVD svd;
	 svd.InvertInPlace(inv);
	 Vector beta_hat;
	 beta_hat.Dimension(trans.fixedEffects);

	 for(int i=0;i<trans.fixedEffects;i++)
	    beta_hat[i] = inv[i].InnerProduct(tmp);

	 Vector residuals;
	 residuals.Dimension(trans.persons);
	 for(int i=0;i<trans.persons;i++)
	    residuals[i] = trans.Y[i] - beta_hat.InnerProduct(trans.X[i]);

	 //double sigma2 = residuals.SumSquare()/trans.persons;

	 //replace trait values into residuals
	 int index = 0;      
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++) 
	    {
	       if (ped[i].isPhenotyped(traitNum) && (!useCovariates || ped[i].isFullyControlled()))
	       {
		  if(PreMeta::genoFromVCF)
		  {
		     int p = trans.sampleVCFIDHash.Integer(ped[i].pid);
		     if(p!=-1)
		     {
			ped[i].traits[traitNum] = residuals[index];
			index++;
		     }
		     else
			ped[i].traits[traitNum] = _NAN_;
		  }
		  if(PreMeta::genoFromPed)
		  {
		     ped[i].traits[traitNum] = residuals[index];
		     index++;
		  }
	       }
	       else
	       {
		  ped[i].traits[traitNum] = _NAN_;
	       }
	    }
	 }
	 useCovariates = false;
      }

      if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
      {
	 if(inverseNormal)
	    InverseNormalTransform(ped,traitNum);

	 fitEngine.useCovariates = FastFit::useCovariates;

	 trans.Prepare(ped,traitNum,useCovariates,log); 

	 fitEngine.FitPolygenicModels(ped,traitNum,true);

	 sigma_g2Hat = fitEngine.heritability * trans.traitVar;
	 sigma_e2Hat = fitEngine.sigma_e2 * trans.traitVar;
	 sigma_gXHat = fitEngine.sigma_gX * trans.traitVar;
	 sigma_s2Hat = fitEngine.sharedEnv * trans.traitVar;
	 betaHat.Copy(fitEngine.betaHat);

	 printf("Model fitting is finished.\n");
	 fprintf(log,"Model fitting is finished.\n");

	 //SVD var-cov matrix and do the transformations
	 trans.TransformPedkinSibEnv(fitEngine, ped);

	 //Fill in UDX, UD, and UDY
	 trans.FinalizeTransform(sigma_g2Hat, sigma_e2Hat);
      }
      else 
      {
	 if(inverseNormal) 
	    InverseNormalTransform(ped,traitNum);

	 trans.Prepare(ped,traitNum,useCovariates,log);

	 if(FastTransform::readInEmp!="" || trans.empKin)
	    trans.TransformEmpkin(ped,traitNum,useCovariates,log);
	 else 
	    trans.TransformPedkin(ped,traitNum,useCovariates,log);

	 OptimizeDelta(0.0001,trans);

	 //Now save the parameter estiamtes
	 GetBeta(deltaHat,trans);
	 betaHat = beta;
	 sigma_g2Hat = GetSigma_g(deltaHat,betaHat,trans);
	 sigma_e2Hat = sigma_g2Hat * deltaHat;
	 printf("Model fitting is finished.\n");
	 fprintf(log,"Model fitting is finished.\n");

	 //Fill in UDX, UD, and UDY
	 trans.FinalizeTransform(sigma_g2Hat, sigma_e2Hat);
      }

      Vector y;
      for (int f = 0; f < ped.familyCount; f++) 
      {    
	 for(int j=0;j<trans.pPheno[f].Length();j++)
	 {
	    y.Push(ped[trans.pPheno[f][j]].traits[traitNum]);
	 }         
      }
      y.Sort();
      int first,second,middle;
      first = (int) y.Length()/4;
      second = first*3;
      middle = (int) (first+second)/2;
      fprintf(SCOREoutput,"##AnalyzedTrait\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",y.Min(),y[first],y[middle],y[second],y.Max(),y.Average(),y.Var());
   }
}

int FastFit::GetTraitID( Pedigree & ped, const char * name,FILE * log)
{
   int idx = ped.traitLookup.Integer(name);
   if (idx != -1) return idx;
   else 
   {
      fprintf(log,"Error! Trait not found. Please check .dat file for the correct trait name.\n");
      error("Error! Trait not found. Please check .dat file for the correct trait name.\n");
      return -1;
   }
}

double FastFit::Brent(double tol,FastTransform & trans)
{
   double temp;

   if (a > c)
   {
      SHIFT(temp, a, c, temp);
      SHIFT(temp, fa, fc, temp);
   }

   min = b; fmin = fb;
   double w = b, v = b;
   double fw = fb, fv = fb;

   double delta = 0.0;         // distance moved in step before last

   double u, fu, d = 0.0;      // Initializing d is not necesary, but avoids warnings
   for (int iter = 1; iter <= ITMAX; iter++)
   {
      double middle = 0.5 * (a + c);
      double tol1 = tol * fabs(min) + ZEPS;
      double tol2 = 2.0 * tol1;

      if (fabs(min - middle) <= (tol2 - 0.5 * (c - a)))
	 return fmin;

      if (fabs(delta) > tol1)
	 // Try a parabolic fit
      {
	 double r = (min - w) * (fmin - fv);
	 double q = (min - v) * (fmin - fw);
	 double p = (min - v) * q - (min - w) * r;

	 q = 2.0 * (q - r);
	 if (q > 0.0) p = -p;
	 q = fabs(q);

	 temp = delta;
	 delta = d;

	 if (fabs(p) >= fabs(0.5 * q * temp) ||
	       p <= q * (a - min) || p >= q * (c - min))
	    // parabolic doesn't look like a good idea
	 {
	    delta = min >= middle ? a - min : c - min;
	    d = CGOLD * delta;
	 }
	 else
	    // parabolic fit is the way to go
	 {
	    d = p / q;
	    u = min + d;
	    if (u - a < tol2 || c - u < tol2)
	       d = sign(tol1, middle - min);
	 }
      }
      else
      {
	 // Golden ratio for first step
	 delta = min >= middle ? a - min : c - min;
	 d = CGOLD * delta;
      }

      // Don't take steps smaller than tol1
      u = fabs(d) >= tol1 ? min + d : min + sign(tol1, d);
      fu = -1.0*Evaluate(u,trans,false);

      if (fu <= fmin)
      {
	 if (u >= min)
	    a = min;
	 else
	    c = min;
	 SHIFT(v, w, min, u);
	 SHIFT(fv, fw, fmin, fu);
      }
      else
      {
	 if (u < min)
	    a = u;
	 else
	    c = u;
	 if (fu <= fw || w == min)
	 {
	    SHFT3(v, w, u);
	    SHFT3(fv, fw, fu);
	 }
	 else if (fu <= fv || v == min || v == w)
	 {
	    v = u;
	    fv = fu;
	 }
      }
   }
   numerror("ScalarMinimizer::Brent got stuck");
   return fmin;
}

void FastFit::EstimateAnalyticalBeta(Vector & Y,FastTransform & trans,Vector & beta_hat)
{
   Vector tmp;
   tmp.Dimension(trans.fixedEffects);
   tmp.Zero();
   for(int i=0;i<trans.persons;i++)
   {
      for(int j=0;j<trans.fixedEffects;j++)
      {
	 tmp[j] += Y[i]*trans.UDX[i][j];
      }
   }
   for(int i=0;i<trans.fixedEffects;i++)
   {
      beta_hat[i] = trans.inv[i].InnerProduct(tmp);
   }
}

/*
   void FastFit::EstimateAnalyticalSigma2(FastTransform & trans,int trait)
   {
   printf("start estimating sigma2...\n");
   for(int p=0;p<trans.persons;p++)
   {
   residuals_orig[trait][p] = trans.UDY[p] -  trans.UDX[p].InnerProduct(analytical_beta_hat[trait]);
   printf("p=%d,",p);
   printf("%g \n",residuals_orig[trait][p]);
   analytical_sigma2_hat[trait] += residuals_orig[trait][p] * residuals_orig[trait][p];
   }
   analytical_sigma2_hat[trait] /= trans.persons;
   printf("a_sigma2 is:%g.\n",analytical_sigma2_hat[trait]);
   }
 */
